<?php

/**
 * @file
 * XeroClient fork for Drupal 7, ext-curl, and ext-json.
 */

/**
 * Xero client helper class.
 */
class XeroClient {

  /**
   * A list of valid tenant guids.
   *
   * @var string[]
   */
  protected $tenantIds = [];

  /**
   * An access token to re-use.
   *
   * @var array
   */
  protected $refreshedToken = NULL;

  /**
   * Request headers keyed by header name.
   *
   * @var array
   */
  protected $defaultHeaders = [];

  /**
   * The base URI to use.
   *
   * @var string
   */
  protected $baseUri = '';

  /**
   * The content-type/response format.
   *
   * @var string
   */
  protected $format = 'xml';

  /**
   * Valid POST methods.
   *
   * @var string[]
   */
  protected $validPostMethods = [
    'BankTransactions', 'BrandingThemes', 'Contacts', 'CreditNotes', 'Invoices',
    'Items', 'ManualJournals', 'Receipts',
  ];

  /**
   * Valid PUT methods.
   *
   * @var string[]
   */
  protected $validPutMethods = [
    'BankTransactions', 'Contacts', 'CreditNotes', 'Employees', 'ExpenseClaims',
    'Invoices', 'Items', 'ManualJournals', 'Payments', 'Receipts',
  ];

  /**
   * Valid GET methods.
   *
   * @var string[]
   */
  protected $validGetMethods = [
    'Accounts', 'BankTransactions', 'BrandingThemes', 'Contacts', 'CreditNotes',
    'Currencies', 'Employees', 'ExpenseClaims', 'Invoices', 'Items', 'Journals',
    'ManualJournals', 'Organisation', 'Payments', 'Receipts', 'TaxRates',
    'TrackingCategories', 'Users',
  ];

  /**
   * Returns valid Xero URLs.
   *
   * @return string[]
   *   A list of valid urls.
   */
  public static function getValidUrls() {
    return [
      'https://identity.xero.com/connect/token',
      'https://api.xero.com/connections',
      'https://api.xero.com/api.xro/2.0/',
      'https://api.xero.com/payroll.xro/1.0/',
      'https://api.xero.com/assets.xro/1.0/',
      'https://api.xero.com/files.xro/1.0/',
    ];
  }

  /**
   * Initialization method.
   *
   * @param string $base_uri
   *   The base URI to use for the request.
   * @param string $auth_token
   *   An authorization token to use.
   * @param string $format
   *   Whether to use json, xml, or pdf. Defaults to xml. Note that a pdf value
   *   will set the accept header, but not the request content-type. To maintain
   *   compatibility, but be less confusing, the default is XML, but all of the
   *   responses will be returned deserialized (or binary data).
   * @param string $tenant_id
   *   An optional tenant id to use.
   */
  public function __construct($base_uri, $auth_token, $format = 'xml', $tenant_id = NULL) {
    if (!$this->isValidUrl($base_uri)) {
      throw new InvalidArgumentException('API URL is not valid.');
    }

    if (!in_array($format, ['json', 'xml', 'pdf'])) {
      throw new InvalidArgumentException('Format is not valid.');
    }

    $this->baseUri = $base_uri;

    $this->defaultHeaders['Authorization'] = 'Bearer ' . $auth_token;
    $this->setFormat($format);

    if ($tenant_id !== NULL) {
      $this->defaultHeaders['xero-tenant-id'] = $tenant_id;
    }
  }

  /**
   * Sets the format to use.
   *
   * @param string $format
   *   The format: xml, json, or pdf.
   */
  public function setFormat($format = 'xml') {
    $this->format = $format;
    $this->defaultHeaders['Content-Type'] = $this->format === 'xml' ? 'text/xml' : 'application/json';
    $this->defaultHeaders['Accept'] = $this->defaultHeaders['Content-Type'];

    if ($this->format === 'pdf') {
      $this->defaultHeaders['Accept'] = 'application/pdf';
    }
  }

  /**
   * Checks if a URL is valid.
   *
   * @param string $base_uri
   *   The url to check.
   *
   * @return bool
   *   TRUE if the URL is a valid Xero URL.
   */
  public function isValidUrl($base_uri) {
    return in_array($base_uri, $this->getValidUrls()) || strpos($base_uri, 'https://api.xero.com/oauth') === 0;
  }

  /**
   * Makes a GET request.
   *
   * @param string $action
   *   The Xero Api Endpoint name or a full URL.
   * @param array $options
   *   Associative array of options to use with the request:
   *     - headers: An associative array keyed by header name.
   *     - params: An associative array keyed by xero property name and value.
   *     - id: An optional Xero GUID to filter the request.
   *
   * @return mixed
   *   The returned data from the API or FALSE.
   */
  public function get($action, array $options = []) {
    $optionalHeaders = isset($options['headers']) ? $options['headers'] : [];
    $params = isset($options['params']) ? $options['params'] : [];
    $query = [];
    $after = isset($options['after']) ? $options['after'] : FALSE;
    $id = isset($options['id']) ? $options['id'] : FALSE;
    $headers = $this->getHeadersFromOptions($optionalHeaders);
    $uri = $this->getUriFromAction($action);

    if (!empty($params)) {
      // Assemble where query parameter.
      $and = '%26%26';
      $query['where'] = '';
      foreach ($params as $key => $value) {
        $where_value = '%3d%3d%22' . $value . '%22';
        if (is_bool($value)) {
          $where_value = ($value) ? '%3d%3dtrue' : '%3d%3dfalse';
        }
        elseif (is_array($value)) {
          if (is_bool($value[1])) {
            $where_value = ($where_value[1])
              ? $value[0] . 'true'
              : $value[0] . 'false';
          }
          else {
            $where_value = $value[0] . '%22{' . $value[1] . '}%22';
          }
        }

        $query['where'] .= strlen($query['where']) === 0
          ? $key . $where_value
          : $and . $key . $where_value;
      }
    }

    if ($after) {
      $headers['If-Modified-Since'] = str_replace('X', 'T',
        date('Y-m-dXH:i:s', $after));
    }

    if ($id) {
      $id = strip_tags(strval($id));
      $uri .= '/' . $id;
    }

    try {
      $res = drupal_http_request($uri, ['data' => $query, 'headers' => $headers]);
      $type = isset($res->headers['content-type']) ? $res->headers['content-type'] : 'application/json';
      $format = strpos($type, $this->format) !== FALSE ? $this->format : 'json';
      $result = $this->unserialize($res->data, $format);

      if (isset($res->error)) {
        if (isset($result['Status'])) {
          xero_error($result['Status'], $result['Type'], $result['Title'], [], $result);
        }
        else {
          xero_error($res->code, 'Error', $res->status_message, [], $res);
        }
      }

      return $result;
    }
    catch (Exception $e) {
      watchdog_exception('xero', $e, 'An error occurred during this request');
      return FALSE;
    }
  }

  /**
   * Makes a put or a post request depending on the type of data.
   *
   * @param string $action
   *   The action to use.
   * @param array $items
   *   The data to send.
   * @param array $options
   *   Extra options to pass in.
   *
   * @return mixed
   *   The response data deserialized or FALSE.
   */
  public function post($action, array $items, array $options = []) {
    if (in_array($action, $this->validPostMethods)) {
      $options['method'] = 'POST';
    }
    elseif (in_array($action, $this->validPutMethods)) {
      $options['method'] = 'PUT';
    }
    else {
      throw new InvalidArgumentException('Action is not valid');
    }

    $options['data'] = $this->serialize($action, $items);
    $options['headers'] = $this->getHeadersFromOptions($options);
    $uri = $this->getUriFromAction($action);

    try {
      $res = drupal_http_request($uri, $options);
      $type = isset($res->headers['content-type']) ? $res->headers['content-type'] : 'application/json';
      $format = strpos($type, $this->format) !== FALSE ? $this->format : 'json';
      $result = $this->unserialize($res->data, $format);

      if (isset($res->error)) {
        if (isset($result['Status'])) {
          xero_error($result['Status'], $result['Type'], $result['Title'], $items, $result);
        }
        else {
          xero_error($res->code, 'Error', $res->status_message, $items, $res);
        }
      }

      return $result;
    }
    catch (Exception $e) {
      watchdog_exception('xero', $e, 'An error occurred with this request');
      return FALSE;
    }
  }

  /**
   * Merges optional headers with default ones.
   *
   * @param array $options
   *   The optional headers.
   *
   * @return array
   *   Associative array of headers keyed by header name.
   */
  protected function getHeadersFromOptions(array $options = []) {
    $headers = isset($options['headers']) ? $options['headers'] : [];
    if (!empty($this->tenantIds) && !isset($options['headers']['xero-tenant-id'])) {
      $headers['xero-tenant-id'] = $this->tenantIds[0]['tenantId'];
    }
    return array_merge($this->defaultHeaders, $headers);
  }

  /**
   * Unserializes data into the format specified.
   *
   * @param string $data
   *   The data from the API.
   * @param string $format
   *   The format.
   *
   * @return array|mixed|string
   *   The unserialized or binary data.
   */
  protected function unserialize($data, $format) {
    if ($format === 'json') {
      return json_decode($data, TRUE);
    }
    elseif ($format === 'xml') {
      $xml = simplexml_load_string($data);
      return XeroXmlHelper::toArray($xml);
    }
    return $data;
  }

  /**
   * Serializes data into the format specified.
   *
   * @param string $type
   *   The plural xero type.
   * @param array $data
   *   The data to send to Xero.
   *
   * @return string
   *   The serialized string.
   */
  protected function serialize($type, array $data) {
    if ($this->format === 'json') {
      return json_encode($data);
    }

    return XeroXmlHelper::fromArray($data, $type);
  }

  /**
   * Get the URI from the action.
   *
   * @param string $action
   *   An action/endpoint or a full URL.
   *
   * @return string
   *   The full URL to use.
   */
  protected function getUriFromAction($action) {
    $scheme = file_uri_scheme($action);
    $uri = $action;

    if (!$scheme) {
      if (!in_array($action, $this->validGetMethods)) {
        throw new \InvalidArgumentException();
      }
      $uri = $this->baseUri . $action;
    }

    return $uri;
  }

  /**
   * Gets the connections that this client is allowed to use.
   *
   * @return array
   *   A list of Xero Connection data.
   */
  public function getConnections() {
    try {
      return $this->get('https://api.xero.com/connections');
    }
    catch (Exception $e) {
      watchdog_exception('xero', $e, 'An error occurred getting connections');
      return [];
    }
  }

  /**
   * Creates a new XeroClient from a valid token.
   *
   * @param string $id
   *   The consumer client identifier.
   * @param string $secret
   *   The consumer secret.
   * @param string $token
   *   A valid access token.
   * @param string $grant
   *   The grant type.
   * @param string $api
   *   The api to use.
   * @param array $options
   *   Options to pass-through such as base uri or xml/json.
   * @param string $redirectUri
   *   The redirect URI to use.
   *
   * @return static
   */
  public static function createFromToken(
    $id,
    $secret,
    $token,
    $grant = NULL,
    $api = 'accounting',
    array $options = [],
    $redirectUri = ''
  ) {
    global $user;

    if ($grant !== NULL) {
      // Fetch a new access token from a refresh token.
      $provider = new \XeroProvider([
        'clientId' => $id,
        'clientSecret' => $secret,
        'scopes' => \XeroProvider::getValidScopes($api),
        'redirectUri' => url('xero/authorize', ['absolute' => TRUE]),
      ]);
      $token_options = [];
      if ($grant === 'refresh_token') {
        $token_options['refresh_token'] = $token;
      }
      elseif ($grant === 'authorization_code') {
        $token_options['code'] = $token;
      }

      $refreshedToken = $provider->getAccessToken($grant, $token_options);
      $token = $refreshedToken['access_token'];
      variable_set('xero_global_token', serialize($refreshedToken));
      module_invoke_all('xero_authorize', $user, $token);
    }

    if (!isset($options['base_uri'])) {
      $options['base_uri'] = 'https://api.xero.com/api.xro/2.0/';
    }

    // Create a new static instance.
    $format = isset($options['format']) ? $options['format'] : 'xml';
    $instance = new static($options['base_uri'], $token, $format);
    $instance->tenantIds = $instance->getConnections();

    if (isset($refreshedToken)) {
      $instance->refreshedToken = $refreshedToken;
    }

    return $instance;
  }

  /**
   * Access tokens refreshed when creating an instance from a refresh token.
   *
   * @return array
   *   The refresh token.
   */
  public function getRefreshedToken() {
    return $this->refreshedToken;
  }

  /**
   * The tenant guids accessible by this client.
   *
   * @return string[]
   *   The guids accessible by this client.
   */
  public function getTenantIds() {
    return $this->tenantIds;
  }

}

/**
 * Serialize and deserialize from XML to Associative Arrays.
 *
 * Xero XML does not contain attributes.
 */
class XeroXmlHelper {

  /**
   * Deserialize to an array.
   *
   * @param \SimpleXMLElement $xml
   *   A simple XML element.
   *
   * @return mixed
   *   The deserialized array or string.
   */
  public static function toArray(SimpleXMLElement $xml) {
    $children = $xml->children();
    if (!$children) {
      return (string) $xml;
    }
    $data = [];

    foreach ($children as $key => $node) {
      // Recurse through each child.
      $childData = self::toArray($node);
      $index = $key === 'anon' ? count($data) : $key;

      if (array_key_exists($index, $data) && isset($data[$index])) {
        // Add the child data to the parent array if it is already set, and if
        // there are multiple of the same node, concatenate it onto the parent
        // array instead.
        if (!is_array($data[$index]) || !isset($data[$index][0]) || (
            array_key_exists(0, $data[$index]) && $data[$index][0] === NULL)) {
          $data[$index] = $childData;
        }
        else {
          $data[$index][] = $childData;
        }
      }
      else {
        $data[$index] = $childData;
      }
    }

    return $data;
  }

  /**
   * Serialize to XML.
   *
   * @param array $data
   *   The associative array to serialize.
   * @param string $pluralRoot
   *   The root node name to use.
   * @param \SimpleXMLElement $xml
   *   The XML document to use. This should not be passed in on the first call.
   *
   * @return string
   *   The xml string. This is ignored in recursed calls.
   */
  public static function fromArray(array $data, string $pluralRoot = 'ResultSet', SimpleXMLElement $xml = NULL) {
    $name = $pluralRoot;
    if ($xml === NULL) {
      $xml = simplexml_load_string("<$pluralRoot />");
      $name = rtrim($pluralRoot, 's');
    }

    foreach ($data as $index => $value) {
      // Create child elements for each array child.
      $key = $index;
      $isNumeric = FALSE;
      if (is_numeric($key)) {
        $key = $name;
        $isNumeric = TRUE;
      }

      $key = preg_replace('/[^a-z0-9_\-\.\:]/i', '', $key);
      if (is_array($value)) {
        // Recurse through the array.
        $node = $isNumeric || self::isAssoc($value) ? $xml->addChild($key) : $xml;
        self::fromArray($value, ($isNumeric ? 'anon' : $key), $node);
      }
      else {
        // Add the non-array element.
        $val = htmlentities($value, ENT_NOQUOTES, 'UTF-8', FALSE);
        $xml->addChild($key, $value);
      }
    }

    return $xml->asXML();
  }

  /**
   * Checks if an array is associative.
   *
   * @param array $toCheck
   *   The array to check.
   *
   * @return bool
   *   TRUE if the array has at least one string key.
   */
  public static function isAssoc(array $toCheck) {
    $stringKeys = array_filter(array_keys($toCheck), function ($key) {
      return !is_int($key);
    });
    return count($stringKeys) > 0;
  }

}
