<?php

/**
 * @file
 * Provides a self-contained Xero OAuth2 provider.
 */

/**
 * A drupal 7 compatible OAuth2 provider for Xero.
 */
class XeroProvider {

  /**
   * Valid scopes.
   *
   * @var string[]
   */
  public static $validScopes = [
    'offline_access', 'openid', 'profile', 'email', 'accounting.transactions',
    'accounting.transactions.read', 'accounting.reports.read',
    'accounting.journals.read', 'accounting.settings', 'accounting.settings.read',
    'accounting.contacts', 'accounting.contacts.read', 'accounting.attachments',
    'accounting.attachments.read', 'payroll.employees', 'payroll.employees.read',
    'payroll.payruns', 'payroll.payruns.read', 'payroll.payslip',
    'payroll.payslip.read', 'payroll.timesheets', 'payroll.timesheets.read',
    'payroll.settings', 'payroll.settings.read', 'files', 'file.read', 'assets',
    'assets.read', 'projects', 'projects.read', 'paymentservices', 'bankfeeds',
  ];

  /**
   * Authorization URL to redirect to.
   *
   * @var string
   */
  public static $baseAuthorizationUrl = 'https://login.xero.com/identity/connect/authorize';

  /**
   * Access token URL.
   *
   * @var string
   */
  public $baseAccessTokenUrl = 'https://identity.xero.com/connect/token';

  /**
   * Stored options from the constructor.
   *
   * @var array
   */
  protected $options = [];

  /**
   * State.
   *
   * @var string
   */
  protected $state;

  /**
   * OAuth client id.
   *
   * @var string
   */
  protected $clientId;

  /**
   * OAuth client secret.
   *
   * @var string
   */
  protected $clientSecret;

  /**
   * The redirect URI to use.
   *
   * @var string
   */
  protected $redirectUri;

  /**
   * Initialization method.
   *
   * @param array $options
   *   An array of options.
   */
  public function __construct(array $options = []) {
    $props = ['state', 'clientId', 'clientSecret', 'redirectUri'];
    $this->options = $options;

    foreach ($props as $prop) {
      $this->{$prop} = isset($options[$prop]) ? $options[$prop] : '';
    }
  }

  /**
   * Gets the state for this request.
   *
   * @return string
   */
  public function getState() {
    return $this->state;
  }

  /**
   * Gets the authorization URL to redirect to.
   *
   * @param array $options
   *   The options to use.
   *
   * @return string
   *   The external URL to redirect to.
   *
   * @throws \Exception
   */
  public function getAuthorizationUrl(array $options) {
    $scopes = isset($options['scope']) ? $options['scope'] : ['offline_access'];
    $url_options = [
      'query' => [
        'response_type' => 'code',
        'client_id' => $this->clientId,
        'redirect_uri' => $this->redirectUri,
        'scope' => implode(' ', $scopes),
        'state' => isset($options['state']) ? $options['state'] : bin2hex(random_bytes(16)),
      ],
      'external' => TRUE,
    ];

    $this->state = $url_options['query']['state'];

    return url(self::$baseAuthorizationUrl, $url_options);
  }

  /**
   * Gets an access token from the Xero API.
   *
   * @param string $grant
   *   The grant type to request.
   * @param array $token_options
   *   Token options to use.
   *
   * @return array
   *   The request token and various metadata.
   */
  public function getAccessToken($grant, array $token_options = []) {
    $params = [];
    $default_params = [
      'grant_type' => $grant,
      'redirect_uri' => $this->redirectUri,
    ];
    $headers = [
      'Content-Type' => 'application/x-www-form-urlencoded',
      'Authorization' => 'Basic ' . base64_encode($this->clientId . ':' . $this->clientSecret),
    ];
    $uri = $this->baseAccessTokenUrl;

    if ($grant === 'authorization_code') {
      $params = array_merge($default_params, ['code' => $token_options['code']]);
    }
    elseif ($grant === 'refresh_token') {
      $params = array_merge($default_params, ['refresh_token' => $token_options['refresh_token']]);
    }

    $res = drupal_http_request($uri, [
      'method' => 'POST',
      'headers' => $headers,
      'data' => $params,
    ]);

    if (!isset($res->error)) {
      return json_decode($res->data, TRUE);
    }
    else {
      watchdog('xero', 'Access token request failed: @data', ['@data' => print_r($res, TRUE)], WATCHDOG_ERROR);
    }

    return [];
  }

  /**
   * Gets the valid scopes for an API.
   *
   * @param string $api
   *   One of openid, accounting, payroll_COUNTRYID, files, assets,
   *   projects, custom, or restricted.
   * @param array $custom
   *   Use the scopes defined here when "custom" is defined above and these
   *   scopes are valid.
   *
   * @return string[]
   *   A list of scopes to use for the api.
   */
  public static function getValidScopes($api = '', array $custom = []) {
    if ($api === 'custom') {
      return array_filter($custom, [__CLASS__, 'isValidScope']);
    }

    $scopes = ['offline_access'];
    if ($api === 'openid') {
      $scopes = array_merge($scopes, ['openid', 'profile', 'email']);
    }
    elseif ($api === 'accounting') {
      $types = ['transactions', 'settings', 'contacts', 'attachments'];
      foreach ($types as $type) {
        $scopes[] = "accounting.$type";
        $scopes[] = "accounting.$type.read";
      }
      $scopes = array_merge($scopes, ['accounting.reports.read', 'accounting.journals.read']);
    }
    elseif (substr($api, 0, 7) === 'payroll') {
      // @todo Split the logic into au, uk, nz, and other sections as necessary.
      $types = ['employees', 'payruns', 'payslip', 'timesheets', 'settings'];
      foreach ($types as $type) {
        $scopes[] = "payroll.$type";
        $scopes[] = "payroll.$type.read";
      }
    }
    elseif ($api === 'files') {
      $scopes = array_merge($scopes, ['files', 'files.read']);
    }
    elseif ($api === 'assets') {
      $scopes = array_merge($scopes, ['assets', 'assets.read']);
    }
    elseif ($api === 'projects') {
      $scopes = array_merge($scopes, ['projects', 'projects.read']);
    }
    elseif ($api === 'restricted') {
      $scopes = array_merge($scopes, ['paymentservices', 'bankfeeds']);
    }

    return $scopes;
  }

  /**
   * Checks if a scope is valid.
   *
   * @param string $scope
   *   The scope to check.
   *
   * @return bool
   *   True if the scope is a valid scope for the Xero API.
   */
  public static function isValidScope($scope) {
    return in_array($scope, static::$validScopes);
  }

}
