<?php

/**
 * @file
 * Xero API administration page callbacks.
 */

/**
 * Administrative settings form callback.
 *
 * @param array $form
 *   The form array.
 * @param array &$form_state
 *   The form state array.
 *
 * @return array
 *   The form array.
 */
function xero_admin_settings($form, &$form_state) {
  $items = [
    [
      'title' => t('Add Xero Application on Xero.com'),
      'href' => 'https://developer.xero.com/myapps/',
    ],
  ];

  $form['help'] = [
    '#theme' => 'links',
    '#links' => $items,
  ];

  // Open the fieldset if there was a problem loading the library.
  $form['xero_consumer_key'] = [
    '#type' => 'textfield',
    '#title' => t('Xero Client ID'),
    '#description' => t('Provide the consumer key for your private application on xero.'),
    '#default_value' => variable_get('xero_consumer_key', ''),
    '#required' => TRUE,
  ];

  $form['xero_consumer_secret'] = [
    '#type' => 'textfield',
    '#title' => t('Xero Client Secret'),
    '#description' => t('Provide the consumer secret for your private application on xero.'),
    '#default_value' => variable_get('xero_consumer_secret', ''),
    '#required' => TRUE,
  ];

  $form['actions'] = ['#type' => 'actions'];
  $form['actions']['submit'] = [
    '#type' => 'submit',
    '#value' => t('Save configuration & Authorize'),
  ];

  return $form;
}

/**
 * Xero admin settings form submit callback.
 *
 * @param array $form
 *   The form array.
 * @param array &$form_state
 *   The form state.
 *
 * @throws \Exception
 */
function xero_admin_settings_submit($form, array &$form_state) {
  $old_key = variable_get('xero_consumer_key', '');
  $old_secret = variable_get('xero_consumer_secret', '');

  // Re-authorize the user.
  variable_set('xero_consumer_key', $form_state['values']['xero_consumer_key']);
  variable_set('xero_consumer_secret', $form_state['values']['xero_consumer_secret']);

  $provider = new XeroProvider([
    'clientId' => $form_state['values']['xero_consumer_key'],
    'clientSecret' => $form_state['values']['xero_consumer_secret'],
    'redirectUri' => url('xero/authorize', ['absolute' => TRUE]),
  ]);

  $form_state['redirect'] = $provider->getAuthorizationUrl([
    'scope' => XeroProvider::getValidScopes('accounting'),
  ]);

  variable_set('xero_temporary_state', $provider->getState());
}

/**
 * Xero default settings form.
 *
 * @param array $form
 *   The form array.
 * @param array &$form_state
 *   The form state array.
 *
 * @return array
 *   The form array.
 */
function xero_admin_default_settings($form, array &$form_state) {
  $account_options = [];
  // Only display defaults if configured correctly.
  $form['defaults'] = [
    '#type' => 'fieldset',
    '#title' => t('Xero API Defaults'),
    '#collapsible' => TRUE,
  ];

  $accounts = xero_get_cache('Accounts');
  if ($accounts !== NULL && !empty($accounts)) {
    foreach ($accounts['Account'] as $account) {
      // Bank accounts do not have a code, exclude them.
      if (!empty($account['Code'])) {
        $account_options[$account['Code']] = $account['Name'];
      }
    }
  }

  $form['defaults']['xero_default_account'] = [
    '#type' => 'select',
    '#title' => t('Default Account'),
    '#description' => t('Choose a default account.'),
    '#options' => $account_options,
    '#default_value' => variable_get('xero_default_account', ''),
  ];

  return system_settings_form($form);
}
